#version 330
#extension GL_ARB_separate_shader_objects : enable





layout(points) in;
layout(location = 0) smooth in vec4 _gin_geomColor[1];
layout(location = 1) smooth in vec2 _gin_geomSize[1];
layout(location = 2) smooth in float _gin_geomRotation[1];
layout(location = 3) smooth in vec3 _gin_geomAxis0[1];
layout(location = 4) smooth in vec3 _gin_geomAxis1[1];

struct SPrimitives
{
	vec4 geomColor;
	vec2 geomSize;
	float geomRotation;
	vec3 geomAxis0;
	vec3 geomAxis1;
	vec4 VertexPosition;
};

struct SGeometryInput
{
	SPrimitives Primitives[1];
	int PrimitiveId;
};

layout(triangle_strip, max_vertices = 6) out;
layout(location = 0) smooth out vec4 _gout_fragColor;

struct SGeometryOutput
{
	vec4 fragColor;
	vec4 VertexPosition;
};

void AppendVertex(in SGeometryOutput outputData)
{
	_gout_fragColor = outputData.fragColor;
	gl_Position = outputData.VertexPosition;
	EmitVertex();
}
void FinishPrimitive(in SGeometryOutput outputData)
{
	EndPrimitive();
}

layout( std140) uniform USceneInfo
{
	mat4 ViewProj;
	mat4 View;
	mat4 Proj;
	mat4 InvView;
	mat4 InvViewProj;
	mat4 UserToLHZ;
	mat4 LHZToUser;
	mat4 BillboardingView;
	mat4 PackNormalView;
	mat4 UnpackNormalView;
	vec4 SideVector;
	vec4 DepthVector;
	vec2 ZBufferLimits;
	float Handedness;
} SceneInfo;

layout( std140) uniform UBillboardInfo
{
	vec4 DrawRequest[1];
} BillboardInfo;
vec2 get_radius(in SPrimitives gInput)
{

	return gInput.geomSize;
}


void rotate_axis(in SPrimitives gInput, inout vec3 xAxis, inout vec3 yAxis)
{

	float c = cos(gInput.geomRotation);
	float s = sin(gInput.geomRotation);
	vec3 xa = xAxis;
	xAxis = yAxis * s + xAxis * c;
	yAxis = yAxis * c - xa * s;

}



void bb_ScreenAlignedQuad(in SPrimitives gInput, out vec3 xAxis, out vec3 yAxis, out vec3 nAxis)
{
	vec2 radius = get_radius(gInput);
	xAxis =  SceneInfo.BillboardingView [0] .xyz;
	yAxis =  SceneInfo.BillboardingView [1] .xyz;
	rotate_axis(gInput, xAxis, yAxis);
	xAxis *= radius.x;
	yAxis *= radius.y;
	nAxis =  SceneInfo.BillboardingView [2] .xyz;
}



void bb_ViewposAlignedQuad(in SPrimitives gInput, out vec3 xAxis, out vec3 yAxis, out vec3 nAxis)
{
	vec3 viewPos =  SceneInfo.BillboardingView [3] .xyz;
	vec3 camToParticle = normalize(gInput.VertexPosition.xyz - viewPos);
	vec3 upVector =  cross( SceneInfo.DepthVector .xyz, SceneInfo.SideVector .xyz) * SceneInfo.Handedness ;
	vec3 upAxis = upVector + 0.01f * dot(upVector, camToParticle) *  SceneInfo.DepthVector .xyz;

	xAxis = normalize( cross(camToParticle, upAxis) * SceneInfo.Handedness );
	yAxis =  cross(xAxis, camToParticle) * SceneInfo.Handedness ;
	rotate_axis(gInput, xAxis, yAxis);
	vec2 radius = get_radius(gInput);
	xAxis *= radius.x;
	yAxis *= radius.y;
	nAxis = -camToParticle;
}



void bb_VelocityAxisAligned(in SPrimitives gInput, out vec3 xAxis, out vec3 yAxis, out vec3 nAxis)
{
	vec3 viewPos =  SceneInfo.BillboardingView [3] .xyz;

	vec3 camToParticle = normalize(gInput.VertexPosition.xyz - viewPos);
	vec2 radius = get_radius(gInput);
	vec3 side =  SceneInfo.SideVector .xyz;
	vec3 depth =  SceneInfo.DepthVector .xyz;

	xAxis =  cross(camToParticle, gInput.geomAxis0) * SceneInfo.Handedness ;
	if (dot(xAxis, xAxis) >  1.0e-8f )
		xAxis = normalize(xAxis);
	else
	{
		vec3 v = -side * dot(camToParticle, depth) + depth * (dot(camToParticle, side) + 0.01f);
		xAxis = normalize( cross(camToParticle, v) * SceneInfo.Handedness );


		xAxis = -xAxis;
	}
	xAxis *= radius.x;

	yAxis = gInput.geomAxis0 * 0.5f;
	nAxis = -camToParticle;
}



void bb_VelocitySpheroidalAlign(in SPrimitives gInput, out vec3 xAxis, out vec3 yAxis, out vec3 nAxis)
{
	vec3 viewPos =  SceneInfo.BillboardingView [3] .xyz;

	vec3 camToParticle = normalize(gInput.VertexPosition.xyz - viewPos);
	vec2 radius = get_radius(gInput);
	vec3 side =  SceneInfo.SideVector .xyz;
	vec3 depth =  SceneInfo.DepthVector .xyz;

	vec3 sideVec =  cross(camToParticle, gInput.geomAxis0) * SceneInfo.Handedness ;
	if (dot(sideVec, sideVec) >  1.0e-8f )
		sideVec = normalize(sideVec);
	else
	{
		vec3 v = -side * dot(camToParticle, depth) + depth * (dot(camToParticle, side) + 0.01f);
		sideVec = normalize( cross(camToParticle, v) * SceneInfo.Handedness );


		sideVec = -sideVec;
	}
	sideVec *= radius.x;

	vec3 upVec =  cross(sideVec, camToParticle) * SceneInfo.Handedness ;
	xAxis = sideVec;
	yAxis = gInput.geomAxis0 * 0.5f + upVec;
	nAxis = -camToParticle;
}



void bb_PlanarAlignedQuad(in SPrimitives gInput, out vec3 xAxis, out vec3 yAxis, out vec3 nAxis)
{
	vec3 center = gInput.VertexPosition.xyz;
	vec3 axis_fwd = gInput.geomAxis0;
	vec3 axis_nrm = gInput.geomAxis1;
	vec3 side =  SceneInfo.SideVector .xyz;
	vec3 depth =  SceneInfo.DepthVector .xyz;

	xAxis =  cross(axis_nrm, axis_fwd) * SceneInfo.Handedness ;
	if (dot(xAxis, xAxis) >  1.0e-8f )
		xAxis = normalize(xAxis);
	else
	{
		vec3 v = -side * dot(axis_nrm, depth) + depth * (dot(axis_nrm, side) + 0.01f);

		xAxis = normalize(cross(axis_nrm, v));


		xAxis = -xAxis;
	}
	yAxis =  cross(xAxis, axis_nrm) * SceneInfo.Handedness ;
	rotate_axis(gInput, xAxis, yAxis);

	vec2 scale = get_radius(gInput);
	xAxis *= scale.x;
	yAxis *= scale.y;
	nAxis = axis_nrm;


	xAxis = -xAxis;
}
void billboard_quad(in SPrimitives gInput, uint bb, out vec3 xAxis, out vec3 yAxis, out vec3 nAxis)
{



	switch (bb)
	{

	case  1U :
		bb_ScreenAlignedQuad(gInput, xAxis, yAxis, nAxis);
		break;


	case  2U :
		bb_ViewposAlignedQuad(gInput, xAxis, yAxis, nAxis);
		break;


	case  3U :
		bb_VelocityAxisAligned(gInput, xAxis, yAxis, nAxis);
		break;


	case  4U :
		bb_VelocitySpheroidalAlign(gInput, xAxis, yAxis, nAxis);
		break;


	case  6U :
		bb_PlanarAlignedQuad(gInput, xAxis, yAxis, nAxis);
		break;

	default:
		bb_ScreenAlignedQuad(gInput, xAxis, yAxis, nAxis);
		break;
	}


	yAxis *= -1.0f;

}

vec4 proj_position(vec3 position)
{

	return  (( SceneInfo.ViewProj ) * (vec4(position, 1.0f))) ;
}

void GeometryMain( in SGeometryInput  gInput, SGeometryOutput gOutput  )
{
	uint drId =  floatBitsToUint(gInput.Primitives[0].VertexPosition.w) ;
	uint flags =  floatBitsToUint( BillboardInfo.DrawRequest [drId].x) ;


	vec3 center = gInput.Primitives[0].VertexPosition.xyz;
	vec3 xAxis = vec3(0, 0, 0);
	vec3 yAxis = vec3(0, 0, 0);
	vec3 nAxis = vec3(0, 0, 0);





	billboard_quad(gInput.Primitives[0], flags &  7U , xAxis, yAxis, nAxis);


	vec3 xpy = xAxis + yAxis;
	vec3 xmy = xAxis - yAxis;
	gOutput.VertexPosition = proj_position(center + xpy); gOutput.fragColor = gInput.Primitives[0].geomColor; AppendVertex(gOutput ); ;
	gOutput.VertexPosition = proj_position(center + xmy); gOutput.fragColor = gInput.Primitives[0].geomColor; AppendVertex(gOutput ); ;
	gOutput.VertexPosition = proj_position(center - xmy); gOutput.fragColor = gInput.Primitives[0].geomColor; AppendVertex(gOutput ); ;
	gOutput.VertexPosition = proj_position(center - xpy); gOutput.fragColor = gInput.Primitives[0].geomColor; AppendVertex(gOutput ); ;

}


void main()
{
	SGeometryInput gInput;
	SGeometryOutput gOutput;

	gInput.Primitives[0].geomColor = _gin_geomColor[0];
	gInput.Primitives[0].geomSize = _gin_geomSize[0];
	gInput.Primitives[0].geomRotation = _gin_geomRotation[0];
	gInput.Primitives[0].geomAxis0 = _gin_geomAxis0[0];
	gInput.Primitives[0].geomAxis1 = _gin_geomAxis1[0];
	gInput.Primitives[0].VertexPosition = gl_in[0].gl_Position;
	gInput.PrimitiveId = gl_PrimitiveIDIn;
	gOutput.fragColor = vec4(0, 0, 0, 0);
	GeometryMain(gInput, gOutput);
}
