#version 330
#extension GL_ARB_separate_shader_objects : enable





layout(location = 0) in vec4 _vin_Position;
layout(location = 1) in vec4 _vin_Color0;
layout(location = 2) in vec4 _vin_Color1;
layout(location = 3) in vec4 _vin_Color2;

struct SVertexInput
{
	vec4 Position;
	vec4 Color0;
	vec4 Color1;
	vec4 Color2;
	int VertexIndex;
	uint InstanceId;
};

layout(location = 0) smooth out vec4 _vout_fragColor0;
layout(location = 1) smooth out vec4 _vout_fragColor1;
layout(location = 2) smooth out vec4 _vout_fragColor2;
layout(location = 3) smooth out vec4 _vout_fragNodeExtendedPixelCoord;
layout(location = 4) smooth out vec4 _vout_fragNodePixelCoord;

struct SVertexOutput
{
	vec4 fragColor0;
	vec4 fragColor1;
	vec4 fragColor2;
	vec4 fragNodeExtendedPixelCoord;
	vec4 fragNodePixelCoord;
	vec4 VertexPosition;
};


layout( std140) uniform UProfilerConstant
{
	mat4 Transform;
} ProfilerConstant;


void VertexMain( in SVertexInput  vInput,  out SVertexOutput  vOutput  )
{
	vec2 curVtxUV;

	switch (vInput.VertexIndex)
	{
		default:
		case 0:
			curVtxUV = vec2(0, 0);
			break;
		case 1:
			curVtxUV = vec2(1, 0);
			break;
		case 2:
			curVtxUV = vec2(1, 1);
			break;
		case 3:
			curVtxUV = vec2(0, 1);
			break;
	}

	vec2 minPos = vInput.Position.xy;
	vec2 maxPos = vInput.Position.zw;



	float currentNodeWidth = maxPos.x - minPos.x;
	float extendNodeSize = 5.0f - currentNodeWidth;
	if (extendNodeSize > 0.0f)
	{
		maxPos.x += extendNodeSize;
		minPos.x -= extendNodeSize;
	}



	vec2 vtxPosition = (vec2(1.0f, 1.0f) - curVtxUV) * minPos + curVtxUV * maxPos;
	vOutput.VertexPosition =  (( ProfilerConstant.Transform ) * (vec4(vtxPosition, 0.0f, 1.0f))) ;


	vOutput.fragColor0 = vInput.Color0;
	vOutput.fragColor1 = vInput.Color1;
	vOutput.fragColor2 = vInput.Color2;

	vOutput.fragNodePixelCoord.xy = vtxPosition - vInput.Position.xy;
	vOutput.fragNodePixelCoord.zw = vInput.Position.zw - vInput.Position.xy;

	vOutput.fragNodeExtendedPixelCoord.xy = vtxPosition - minPos;
	vOutput.fragNodeExtendedPixelCoord.zw = maxPos - minPos;
}


void main()
{
	SVertexInput vInput;
	SVertexOutput vOutput;

	vInput.Position = _vin_Position;
	vInput.Color0 = _vin_Color0;
	vInput.Color1 = _vin_Color1;
	vInput.Color2 = _vin_Color2;
	vInput.VertexIndex = gl_VertexID;
	vInput.InstanceId = uint(gl_InstanceID);
	vOutput.fragColor0 = vec4(0, 0, 0, 0);
	vOutput.fragColor1 = vec4(0, 0, 0, 0);
	vOutput.fragColor2 = vec4(0, 0, 0, 0);
	vOutput.fragNodeExtendedPixelCoord = vec4(0, 0, 0, 0);
	vOutput.fragNodePixelCoord = vec4(0, 0, 0, 0);
	VertexMain(vInput, vOutput);
	_vout_fragColor0 = vOutput.fragColor0;
	_vout_fragColor1 = vOutput.fragColor1;
	_vout_fragColor2 = vOutput.fragColor2;
	_vout_fragNodeExtendedPixelCoord = vOutput.fragNodeExtendedPixelCoord;
	_vout_fragNodePixelCoord = vOutput.fragNodePixelCoord;
	gl_Position = vOutput.VertexPosition;
}
